# -*- coding: utf-8 -*-
# Elisa - Home multimedia server
# Copyright (C) 2006-2008 Fluendo Embedded S.L. (www.fluendo.com).
# All rights reserved.
#
# This file is available under one of two license agreements.
#
# This file is licensed under the GPL version 3.
# See "LICENSE.GPL" in the root of this distribution including a special
# exception to use Elisa with Fluendo's plugins.
#
# The GPL part of Elisa is also available under a commercial licensing
# agreement from Fluendo.
# See "LICENSE.Elisa" in the root directory of this distribution package
# for details on that license.
#
# Author: Olivier Tilloy <olivier@fluendo.com>

"""
Widget and controller for a popup dialog.
"""

from elisa.core.input_event import EventValue

from elisa.plugins.pigment.pigment_controller import PigmentController
from elisa.plugins.pigment.widgets.widget import Widget
from elisa.plugins.pigment.widgets.theme import Theme
from elisa.plugins.pigment.graph.image import Image
from elisa.plugins.pigment.graph.text import Text

from elisa.plugins.poblesec.widgets.sliced_image import SlicedImageHorizontal
from elisa.plugins.poblesec.widgets.button import TextButton

import pgm
from pgm.timing import implicit


class PopupButton(TextButton):
    style_prefix = 'popup-button'


class PopupWidget(Widget):

    """
    Popup widget (containing an icon, a title, a body text, and buttons).
    """

    style_prefix = 'popup'

    def __init__(self, icon, title, text, buttons):
        super(PopupWidget, self).__init__()

        self.background = Image()
        self.add(self.background)
        self.background.position = (0.0, 0.0, 0.0)
        self.background.size = (1.0, 1.0)
        self.background.visible = True

        self._icon = Image()
        self.add(self._icon)
        theme = Theme.get_default()
        image = theme.get_resource(icon)
        self._icon.set_from_file(image)
        self._icon.bg_a = 0
        self._icon.visible = True

        self._title = Text()
        self.add(self._title)
        self._title.label = title
        self._title.bg_a = 0
        self._title.visible = True

        self._text = Text()
        self.add(self._text)
        self._text.label = text
        self._text.bg_a = 0
        self._text.visible = True

        self.buttons = buttons
        self._buttons = []
        self.callbacks = {}

        def button_clicked_cb(button_widget, x, y, z, button, time, pressure):
            self.selected_item_index = self._buttons.index(button_widget)
            self.callbacks[button_widget]()
            return True

        for text, callback in self.buttons:
            button = PopupButton()
            self._buttons.append(button)
            self.add(button)
            button.label = text
            button.bg_a = 0
            button.visible = True

            self.callbacks[button] = callback
            button.connect('clicked', button_clicked_cb)

        self._selected_item_index = -1
        self._selector = SlicedImageHorizontal()
        self.add(self._selector)
        resource_root = 'elisa.plugins.poblesec.selector'
        left_cap = theme.get_resource('%s_left_cap' % resource_root)
        right_cap = theme.get_resource('%s_right_cap' % resource_root)
        body = theme.get_resource('%s_body' % resource_root)
        self._selector.left_cap.set_from_file(left_cap)
        self._selector.right_cap.set_from_file(right_cap)
        self._selector.body.set_from_file(body)
        self._selector.visible = True

        self._animated_selector = implicit.AnimatedObject(self._selector)
        self._animated_selector.setup_next_animations(duration=200)

        self._update_style_properties(self._style.get_items())

    def _update_style_properties(self, props=None):
        super(PopupWidget, self)._update_style_properties(props)

        if props is None:
            return

        for key, value in props.iteritems():
            if key == '%s-bg-color' % self.style_prefix:
                self.background.bg_color = value

            elif key == '%s-icon-x' % self.style_prefix:
                self._icon.x = value
            elif key == '%s-icon-y' % self.style_prefix:
                self._icon.y = value
            elif key == '%s-icon-z' % self.style_prefix:
                self._icon.z = value
            elif key == '%s-icon-width' % self.style_prefix:
                self._icon.width = value
            elif key == '%s-icon-height' % self.style_prefix:
                self._icon.height = value

            elif key == '%s-title-x' % self.style_prefix:
                self._title.x = value
            elif key == '%s-title-y' % self.style_prefix:
                self._title.y = value
            elif key == '%s-title-z' % self.style_prefix:
                self._title.z = value
            elif key == '%s-title-width' % self.style_prefix:
                self._title.width = value
            elif key == '%s-title-height' % self.style_prefix:
                self._title.height = value
            elif key == '%s-title-weight' % self.style_prefix:
                self._title.weight = eval('pgm.TEXT_WEIGHT_%s' % value.upper())
            elif key == '%s-title-fg-color' % self.style_prefix:
                self._title.fg_color = value

            elif key == '%s-text-x' % self.style_prefix:
                self._text.x = value
            elif key == '%s-text-y' % self.style_prefix:
                self._text.y = value
            elif key == '%s-text-z' % self.style_prefix:
                self._text.z = value
            elif key == '%s-text-width' % self.style_prefix:
                self._text.width = value
            elif key == '%s-text-height' % self.style_prefix:
                self._text.height = value
            elif key == '%s-text-weight' % self.style_prefix:
                self._text.weight = eval('pgm.TEXT_WEIGHT_%s' % value.upper())
            elif key == '%s-text-fg-color' % self.style_prefix:
                self._text.fg_color = value
            elif key == '%s-text-font-height' % self.style_prefix:
                self._text.font_height = value

            elif key == '%s-buttons-x' % self.style_prefix:
                for index, button in enumerate(self._buttons):
                    button.x = value + \
                        index * props['%s-button-width' % self.style_prefix]
                self._selector.x = value
            elif key == '%s-buttons-y' % self.style_prefix:
                for button in self._buttons:
                    button.y = value
                self._selector.y = value

            elif key == '%s-button-width' % self.style_prefix:
                for button in self._buttons:
                    button.width = value
                self._selector.width = value
            elif key == '%s-button-height' % self.style_prefix:
                for button in self._buttons:
                    button.height = value
                self._selector.height = value

            elif key == '%s-selector-bg-color' % self.style_prefix:
                self._selector.bg_color = value

    def clean(self):
        self._animated_selector = None
        self._selector.clean()
        self._selector = None
        self._text.clean()
        self._text = None
        self._title.clean()
        self._title = None
        self._icon.clean()
        self._icon = None
        self.background.clean()
        self.background = None
        return super(PopupWidget, self).clean()

    def _set_selected_item_index(self, value):
        if value >= 0 and value < len(self._buttons):
            self._selected_item_index = value
            button = self._buttons[value]
            button.focus = True
            self._animated_selector.x = button.x

    def _get_selected_item_index(self):
        return self._selected_item_index

    selected_item_index = property(fget=_get_selected_item_index,
                                   fset=_set_selected_item_index)


class PopupController(PigmentController):

    """
    Popup controller.
    """

    def initialize(self, icon, title, text, buttons):
        dfr = super(PopupController, self).initialize()
        self.icon = icon
        self.title = title
        self.text = text
        self.buttons = buttons
        dfr.addCallback(self._create_widget)
        dfr.addCallback(self._connect_to_signals)
        return dfr

    def _create_widget(self, result):
        self.popup = PopupWidget(self.icon, self.title,
                                 self.text, self.buttons)
        self.widget.add(self.popup)
        self.popup.position = (0.0, 0.0, 0.1)
        self.popup.size = (1.0, 1.0)
        self.popup.visible = True
        return result

    def _connect_to_signals(self, result):
        self._mouse_event_handler_ids = []
        for event in ['double-clicked', 'drag-end', 'released', 'clicked',
                      'pressed', 'drag-begin', 'drag-motion']:
            event_id = self.popup.background.connect(event,
                                                     self._mouse_event_cb)
            self._mouse_event_handler_ids.append(event_id)

        scrolled_event_id = self.popup.background.connect('scrolled',
                                                          self._scrolled_cb)
        self._mouse_event_handler_ids.append(scrolled_event_id)

        # By default the focus is forwarded to the first button of the popup
        self._focus_handler_id = self.widget.connect('focus', self._on_focus)

        return result

    def clean(self):
        self.widget.disconnect(self._focus_handler_id)
        for event_id in self._mouse_event_handler_ids:
            self.popup.background.disconnect(event_id)
        self.popup.clean()
        self.popup = None
        return super(PopupController, self).clean()

    def _mouse_event_cb(self, *args):
        # Swallow the event
        return True

    def _scrolled_cb(self, widget, x, y, z, direction, time):
        if direction == pgm.SCROLL_DOWN:
            self.popup.selected_item_index -= 1
        elif direction == pgm.SCROLL_UP:
            self.popup.selected_item_index += 1
        return True

    def _on_focus(self, widget, focus):
        # always forward the focus to the buttons of the popup
        if focus:
            self.popup.selected_item_index = 0

    def handle_input(self, manager, input_event):
        if input_event.value in (EventValue.KEY_OK, EventValue.KEY_RETURN):
            self.buttons[self.popup.selected_item_index][1]()
        elif input_event.value == EventValue.KEY_GO_LEFT:
            self.popup.selected_item_index -= 1
        elif input_event.value == EventValue.KEY_GO_RIGHT:
            self.popup.selected_item_index += 1
        elif input_event.value in (EventValue.KEY_f, EventValue.KEY_EXIT):
            # Propagate only the 'fullscreen' and 'exit' events. All the other
            # events are swallowed because the popup is modal.
            su = super(PopupController, self)
            return su.handle_input(manager, input_event)
        return True
